import asyncio
import io
import re
import typing

import discord

from oobabot import fancy_logger
from oobabot import sd_client
from oobabot import templates


async def image_task_to_file(image_task: "asyncio.Task[bytes]", image_request: str):
    await image_task
    img_bytes = image_task.result()
    file_of_bytes = io.BytesIO(img_bytes)
    file = discord.File(file_of_bytes)
    file.filename = "photo.png"
    file.description = f"image generated from '{image_request}'"
    return file


class StableDiffusionImageView(discord.ui.View):
    """
    A View that displays buttons to regenerate an image
    from Stable Diffusion with a new seed, or to lock
    in the current image.
    """

    def __init__(
        self,
        sd_client: sd_client.StableDiffusionClient,
        is_channel_nsfw: bool,
        image_prompt: str,
        requesting_user_id: int,
        requesting_user_name: str,
        template_store: templates.TemplateStore,
    ):
        super().__init__(timeout=120.0)

        self.template_store = template_store

        # only the user who requested generation of the image
        # can have it replaced
        self.requesting_user_id = requesting_user_id
        self.requesting_user_name = requesting_user_name
        self.image_prompt = image_prompt
        self.photo_accepted = False

        #####################################################
        # "Try Again" button
        #
        btn_try_again = discord.ui.Button(
            label="Try Again",
            style=discord.ButtonStyle.blurple,
            row=1,
        )
        self.image_message = None

        async def on_try_again(interaction: discord.Interaction):
            result = await self.diy_interaction_check(interaction)
            if not result:
                # unauthorized user
                return

            try:
                # generate a new image
                regen_task = sd_client.generate_image(image_prompt, is_channel_nsfw)
                regen_file = await image_task_to_file(regen_task, image_prompt)
                await interaction.response.defer()
                await self.get_image_message().edit(attachments=[regen_file])
            except Exception as e:
                fancy_logger.get().error(f"Could not regenerate image: {e}")

        btn_try_again.callback = on_try_again

        #####################################################
        # "Accept" button
        #
        btn_lock_in = discord.ui.Button(
            label="Accept",
            style=discord.ButtonStyle.success,
            row=1,
        )

        async def on_lock_in(interaction: discord.Interaction):
            result = await self.diy_interaction_check(interaction)
            if not result:
                # unauthorized user
                return
            await interaction.response.defer()
            await self.detach_view_keep_img()

        btn_lock_in.callback = on_lock_in

        #####################################################
        # "Delete" button
        #
        btn_delete = discord.ui.Button(
            label="Delete",
            style=discord.ButtonStyle.danger,
            row=1,
        )

        async def on_delete(interaction: discord.Interaction):
            result = await self.diy_interaction_check(interaction)
            if not result:
                # unauthorized user
                return
            await interaction.response.defer()
            await self.delete_image()

        btn_delete.callback = on_delete

        super().add_item(btn_try_again).add_item(btn_lock_in).add_item(btn_delete)

    def set_image_message(self, image_message: discord.Message):
        self.image_message = image_message

    def get_image_message(self) -> discord.Message:
        if self.image_message is None:
            raise ValueError("image_message is None")
        return self.image_message

    async def delete_image(self):
        await self.detach_view_delete_img(self.get_detach_message())

    async def detach_view_delete_img(self, detach_msg: str):
        await self.get_image_message().edit(
            content=detach_msg,
            view=None,
            attachments=[],
        )

    async def detach_view_keep_img(self):
        self.photo_accepted = True
        await self.get_image_message().edit(
            content=None,
            view=None,
        )

    async def on_timeout(self):
        if not self.photo_accepted:
            await self.delete_image()

    async def diy_interaction_check(self, interaction: discord.Interaction) -> bool:
        """
        Only allow the requesting user to interact with this view.
        """
        if interaction.user.id == self.requesting_user_id:
            return True
        error_mesage = self._get_message(templates.Templates.IMAGE_UNAUTHORIZED)
        await interaction.response.send_message(
            content=error_mesage,
            ephemeral=True,
        )
        return False

    def get_image_message_text(self) -> str:
        return self._get_message(templates.Templates.IMAGE_CONFIRMATION)

    def get_detach_message(self) -> str:
        return self._get_message(templates.Templates.IMAGE_DETACH)

    def _get_message(self, message_type: templates.Templates) -> str:
        return self.template_store.format(
            message_type,
            {
                templates.TemplateToken.USER_NAME: self.requesting_user_name,
                templates.TemplateToken.IMAGE_PROMPT: self.image_prompt,
            },
        )


class ImageGenerator:
    def __init__(
        self,
        stable_diffusion_client: sd_client.StableDiffusionClient,
        image_words: typing.List[str],
        template_store: templates.TemplateStore,
    ):
        self.stable_diffusion_client = stable_diffusion_client
        self.template_store = template_store
        self.image_patterns = [
            re.compile(
                r"^.*\b" + image_word + r"\b[\s]*(of|with)?[\s]*[:]?(.*)$",
                re.IGNORECASE,
            )
            for image_word in image_words
        ]

    async def _generate_image(
        self,
        image_prompt: str,
        raw_message: discord.Message,
        response_channel: discord.abc.Messageable,
    ) -> discord.Message:
        is_channel_nsfw = False

        # note: public threads in NSFW chanels are not considered here
        if isinstance(raw_message.channel, discord.TextChannel):
            is_channel_nsfw = raw_message.channel.is_nsfw()

        image_task = self.stable_diffusion_client.generate_image(
            image_prompt, is_channel_nsfw=is_channel_nsfw
        )
        try:
            file = await image_task_to_file(image_task, image_prompt)
        except Exception as e:
            fancy_logger.get().error(f"Could not generate image: {e}")
            error_message = self.template_store.format(
                templates.Templates.IMAGE_GENERATION_ERROR,
                {
                    templates.TemplateToken.USER_NAME: raw_message.author.display_name,
                    templates.TemplateToken.IMAGE_PROMPT: image_prompt,
                },
            )
            return await response_channel.send(error_message, reference=raw_message)

        regen_view = StableDiffusionImageView(
            self.stable_diffusion_client,
            is_channel_nsfw=is_channel_nsfw,
            image_prompt=image_prompt,
            requesting_user_id=raw_message.author.id,
            requesting_user_name=raw_message.author.display_name,
            template_store=self.template_store,
        )

        kwargs = {}
        # we can only pass a reference if the message is in the same channel
        # as the original request.  Also, send() won't take None of this
        # argument, so we need to conditionally add it.
        if raw_message.channel == response_channel:
            kwargs["reference"] = raw_message

        image_message = await response_channel.send(
            content=regen_view.get_image_message_text(),
            file=file,
            view=regen_view,
            **kwargs,
        )
        regen_view.image_message = image_message
        return image_message

    def maybe_get_image_prompt(
        self, raw_message: discord.Message
    ) -> typing.Optional[str]:
        image_prompt = None
        for image_pattern in self.image_patterns:
            match = image_pattern.search(raw_message.content)
            if match:
                image_prompt = match.group(2)
                return image_prompt
        if image_prompt is None:
            return None

        fancy_logger.get().debug("Found image prompt: %s", image_prompt)

    async def generate_image(
        self,
        image_prompt: str,
        raw_message: discord.Message,
        response_channel: discord.abc.Messageable,
    ) -> typing.Optional["asyncio.Task[discord.Message]"]:
        """
        If the message contains a photo word, kick off a task
        to generate an image, post it to the channel, and return
        the generated message.

        If the message does not contain a photo word, return None.
        """
        create_image_task = asyncio.create_task(
            self._generate_image(image_prompt, raw_message, response_channel)
        )
        return create_image_task



# purpose: generate a prompt for the AI to respond to, given
# the message history and persona.
import typing

from oobabot import fancy_logger
from oobabot import templates
from oobabot import types


class PromptGenerator:
    """
    Purpose: generate a prompt for the AI to use, given
    the message history and persona.
    """

    # this is set by the AI, and is the maximum length
    # it will understand before it starts to ignore
    # the rest of the prompt_prefix
    # note: we don't currently measure tokens, we just
    # count characters. This is a rough estimate.
    EST_CHARACTERS_PER_TOKEN = 3

    # the estimated number of characters in a line of message history
    # this is used to rougly calculate whether we'll have enough space
    # to supply the requested number of lines of history.
    #
    # in practice, we will look at the actual number of characters to
    # see what we can fit.
    #
    # note that we're doing calculations in characters, not in tokens,
    # so even counting characters exactly is still an estimate.
    EST_CHARACTERS_PER_HISTORY_LINE = 30

    # when we're not splitting responses, each history line is
    # much larger, and it's easier to run out of token space,
    # so we use a different estimate
    EST_CHARACTERS_PER_HISTORY_LINE_NOT_SPLITTING_RESPONSES = 180

    def __init__(
        self,
        ai_name: str,
        persona: str,
        history_lines: int,
        token_space: int,
        template_store: templates.TemplateStore,
        dont_split_responses: bool,
    ):
        self.ai_name = ai_name
        self.persona = persona
        self.history_lines = history_lines
        self.token_space = token_space
        self.template_store = template_store
        self.dont_split_responses = dont_split_responses

        # this will be also used when sending message
        # to suppress sending the prompt text to the user
        self.bot_prompt_line = self.template_store.format(
            templates.Templates.PROMPT_HISTORY_LINE,
            {
                templates.TemplateToken.USER_NAME: self.ai_name,
                templates.TemplateToken.USER_MESSAGE: "",
            },
        ).strip()

        self.image_request_made = self.template_store.format(
            templates.Templates.PROMPT_IMAGE_COMING,
            {
                templates.TemplateToken.AI_NAME: self.ai_name,
            },
        )
        self._init_history_available_chars()

    def _init_history_available_chars(self) -> None:
        """
        Calculate the number of characters we have available
        for history, and raise an exception if we don't have
        enough.

        Raises:
            ValueError: if we don't estimate to have enough space
                for the requested number of lines of history
        """
        # the number of chars we have available for history
        # is:
        #   number of chars in token space (estimated)
        #   minus the number of chars in the prompt
        #     - without any history
        #     - but with the photo request
        #
        est_chars_in_token_space = self.token_space * self.EST_CHARACTERS_PER_TOKEN
        prompt_without_history = self._generate("", self.image_request_made)

        # how many chars might we have available for history?
        available_chars_for_history = est_chars_in_token_space - len(
            prompt_without_history
        )
        # how many chars do we need for the requested number of
        # lines of history?
        chars_per_history_line = self.EST_CHARACTERS_PER_HISTORY_LINE
        if self.dont_split_responses:
            chars_per_history_line = (
                self.EST_CHARACTERS_PER_HISTORY_LINE_NOT_SPLITTING_RESPONSES
            )

        required_history_size_chars = self.history_lines * chars_per_history_line

        if available_chars_for_history < required_history_size_chars:
            raise ValueError(
                "AI token space is too small for prompt_prefix and history "
                + "by an estimated "
                + f"{required_history_size_chars - available_chars_for_history}"
                + " characters.  You may lose history context.  You can save space"
                + " by shortening the persona or reducing the requested number of"
                + " lines of history."
            )
        self.max_history_chars = available_chars_for_history

    async def _render_history(
        self,
        ai_user_id: int,
        message_history: typing.AsyncIterator[types.GenericMessage],
        stop_before_message_id: typing.Optional[int],
    ) -> str:
        # add on more history, but only if we have room
        # if we don't have room, we'll just truncate the history
        # by discarding the oldest messages first
        # this is s
        # it will understand before ignore
        #
        prompt_len_remaining = self.max_history_chars

        # history_lines is newest first, so figure out
        # how many we can take, then append them in
        # reverse order
        history_lines = []

        async for message in message_history:
            # if we've hit the throttle message, stop and don't add any
            # more history
            if stop_before_message_id and message.message_id == stop_before_message_id:
                break

            adjusted_author_name = message.author_name
            if message.author_id == ai_user_id:
                # make sure the AI always sees its persona name
                # in the transcript, even if the chat program
                # has it under a different account name
                adjusted_author_name = self.ai_name

                # hack: if the message includes the text
                # "tried to make an image with the prompt",
                # ignore it
                if "tried to make an image with the prompt" in message.body_text:
                    continue

            if not message.body_text:
                continue

            line = self.template_store.format(
                templates.Templates.PROMPT_HISTORY_LINE,
                {
                    templates.TemplateToken.USER_NAME: adjusted_author_name,
                    templates.TemplateToken.USER_MESSAGE: message.body_text,
                },
            )

            if len(line) > prompt_len_remaining:
                num_discarded_lines = self.history_lines - len(history_lines)
                fancy_logger.get().warn(
                    "ran out of prompt space, discarding "
                    + f"{num_discarded_lines} lines "
                    + "of chat history"
                )
                break

            prompt_len_remaining -= len(line)
            history_lines.append(line)

        history_lines.reverse()
        return "".join(history_lines)

    def _generate(
        self,
        message_history_txt: str,
        image_coming: str,
    ) -> str:
        prompt = self.template_store.format(
            templates.Templates.PROMPT,
            {
                templates.TemplateToken.AI_NAME: self.ai_name,
                templates.TemplateToken.PERSONA: self.persona,
                templates.TemplateToken.MESSAGE_HISTORY: message_history_txt,
                templates.TemplateToken.IMAGE_COMING: image_coming,
            },
        )
        # todo: make this part of the template?
        prompt += self.bot_prompt_line + "\n"
        return prompt

    async def generate(
        self,
        ai_user_id: int,
        message_history: typing.Optional[typing.AsyncIterator[types.GenericMessage]],
        image_requested: bool,
        throttle_message_id: int,
    ) -> str:
        """
        Generate a prompt for the AI to respond to.
        """
        message_history_txt = ""
        if message_history is not None:
            message_history_txt = await self._render_history(
                ai_user_id, message_history, throttle_message_id
            )
        image_coming = self.image_request_made if image_requested else ""
        return self._generate(message_history_txt, image_coming)



import re
import typing

import discord

from oobabot import fancy_logger
from oobabot import types

FORBIDDEN_CHARACTERS = r"[\n\r\t]"
FORBIDDEN_CHARACTERS_PATTERN = re.compile(FORBIDDEN_CHARACTERS)


def get_channel_name(channel: discord.abc.Messageable) -> str:
    if isinstance(channel, discord.Thread):
        return "thread #" + channel.name
    if isinstance(channel, discord.abc.GuildChannel):
        return "channel #" + channel.name
    if isinstance(channel, discord.DMChannel):
        return "-DM-"
    if isinstance(channel, discord.GroupChannel):
        return "-GROUP-DM-"
    return "-Unknown-"


def sanitize_string(raw_string: str) -> str:
    """
    Filter out any characters that would confuse the AI
    """
    return FORBIDDEN_CHARACTERS_PATTERN.sub(" ", raw_string)


def discord_message_to_generic_message(
    raw_message: discord.Message,
) -> typing.Union[types.GenericMessage, types.ChannelMessage, types.DirectMessage]:
    """
    Convert a discord message to a GenericMessage or subclass thereof
    """
    generic_args = {
        "author_id": raw_message.author.id,
        "author_name": sanitize_string(raw_message.author.display_name),
        "channel_id": raw_message.channel.id,
        "channel_name": get_channel_name(raw_message.channel),
        "message_id": raw_message.id,
        "body_text": sanitize_string(raw_message.content),
        "author_is_bot": raw_message.author.bot,
        "send_timestamp": raw_message.created_at.timestamp(),
    }
    if isinstance(raw_message.channel, discord.DMChannel):
        return types.DirectMessage(**generic_args)
    if (
        isinstance(raw_message.channel, discord.TextChannel)
        or isinstance(raw_message.channel, discord.GroupChannel)
        or isinstance(raw_message.channel, discord.Thread)
    ):
        return types.ChannelMessage(
            mentions=[mention.id for mention in raw_message.mentions],
            **generic_args,
        )
    fancy_logger.get().warning(
        f"Unknown channel type {type(raw_message.channel)}, "
        + f"unsolicited replies disabled.: {raw_message.channel}"
    )
    return types.GenericMessage(**generic_args)



# Purpose: Discord client for Rosie
#

import asyncio
import typing

import discord

from oobabot import bot_commands
from oobabot import decide_to_respond
from oobabot import discord_utils
from oobabot import fancy_logger
from oobabot import image_generator
from oobabot import ooba_client
from oobabot import prompt_generator
from oobabot import repetition_tracker
from oobabot import response_stats
from oobabot import types


class DiscordBot(discord.Client):
    # seconds after which we'll lazily purge a channel
    # from channel_last_direct_response

    def __init__(
        self,
        ooba_client: ooba_client.OobaClient,
        decide_to_respond: decide_to_respond.DecideToRespond,
        prompt_generator: prompt_generator.PromptGenerator,
        repetition_tracker: repetition_tracker.RepetitionTracker,
        response_stats: response_stats.AggregateResponseStats,
        bot_commands: bot_commands.BotCommands,
        image_generator: typing.Optional[image_generator.ImageGenerator],
        ai_name: str,
        persona: str,
        ignore_dms: bool,
        dont_split_responses: bool,
        reply_in_thread: bool,
        log_all_the_things: bool,
    ):
        self.ooba_client = ooba_client
        self.decide_to_respond = decide_to_respond
        self.prompt_generator = prompt_generator
        self.repetition_tracker = repetition_tracker
        self.response_stats = response_stats

        self.ai_name = ai_name
        self.persona = persona
        self.ai_user_id = -1
        self.image_generator = image_generator

        self.ignore_dms = ignore_dms
        self.dont_split_responses = dont_split_responses
        self.reply_in_thread = reply_in_thread
        self.log_all_the_things = log_all_the_things

        # a list of timestamps in which we last posted to a channel
        self.channel_last_direct_response = {}

        self.bot_commands = bot_commands

        intents = discord.Intents.default()
        intents.message_content = True
        intents.members = True

        super().__init__(intents=intents)

    async def on_ready(self) -> None:
        guilds = self.guilds
        num_guilds = len(guilds)
        num_channels = sum([len(guild.channels) for guild in guilds])

        if self.user:
            self.ai_user_id = self.user.id
            user_id_str = str(self.ai_user_id)
        else:
            user_id_str = "<unknown>"

        fancy_logger.get().info(
            f"Connected to discord as {self.user} (ID: {user_id_str})"
        )
        fancy_logger.get().debug(
            f"monitoring {num_channels} channels across " + f"{num_guilds} server(s)"
        )
        if self.ignore_dms:
            fancy_logger.get().debug("Ignoring DMs")
        else:
            fancy_logger.get().debug("listening to DMs")

        if self.dont_split_responses:
            fancy_logger.get().debug("Responses: returned as single messages")
        else:
            fancy_logger.get().debug("Responses: streamed as separate sentences")

        if self.image_generator:
            fancy_logger.get().debug("Image generation: enabled")
        else:
            fancy_logger.get().debug("Image generation: disabled")

        fancy_logger.get().debug(f"AI name: {self.ai_name}")
        fancy_logger.get().debug(f"AI persona: {self.persona}")

        fancy_logger.get().debug(
            f"History: {self.prompt_generator.history_lines} lines "
        )

        str_wakewords = (
            ", ".join(self.decide_to_respond.wakewords)
            if self.decide_to_respond.wakewords
            else "<none>"
        )
        fancy_logger.get().debug(f"Wakewords: {str_wakewords}")

        # we do this at the very end because when you restart
        # the bot, it can take a while for the commands to
        # register
        try:
            # register the commands
            await self.bot_commands.on_ready(self)
        except Exception as e:
            fancy_logger.get().warning(
                f"Failed to register commands: {e} (continuing without commands)"
            )

    async def on_message(self, raw_message: discord.Message) -> None:
        try:
            message = discord_utils.discord_message_to_generic_message(raw_message)
            should_respond, is_summon = self.decide_to_respond.should_reply_to_message(
                self.ai_user_id, message
            )
            if not should_respond:
                return

            async with raw_message.channel.typing():
                await self._handle_response(message, raw_message, is_summon)

        except Exception as e:
            fancy_logger.get().error(
                f"Exception while processing message: {e}", exc_info=True
            )

    async def _handle_response(
        self,
        message: types.GenericMessage,
        raw_message: discord.Message,
        is_summon: bool,
    ):
        image_prompt = None
        if self.image_generator is not None:
            # are we creating an image?
            image_prompt = self.image_generator.maybe_get_image_prompt(raw_message)

        result = await self.send_response(
            message=message,
            raw_message=raw_message,
            image_requested=image_prompt is not None,
        )
        if result is None:
            # we failed to create a thread that the user could
            # read our response in, so we're done here.  Abort!
            return
        message_task, response_channel = result

        # log the mention, now that we know the channel we
        # want to montior later to continue to conversation
        if isinstance(message, types.ChannelMessage):
            # this logic is weird, so let's explain it...
            #
            # we're eventually going to call log_mention()
            # if all these conditions pass.  When we do this,
            # we'll start monitoring the channel_id in the near
            # future for replies that we might respond to, unprompted.
            #
            # In the general case, we want to monitor the channel
            # only if we were summoned in it.
            #
            # However if we were summonned in a channel but are
            # creating a new thread for the answer (because of
            # the --reply-in-thread flag), we want to monitor
            # that thread, not the original channel.
            #
            if is_summon:
                if isinstance(response_channel, discord.Thread):
                    message.channel_id = response_channel.id
                self.decide_to_respond.log_mention(message)

        image_task = None
        if self.image_generator is not None and image_prompt is not None:
            image_task = await self.image_generator.generate_image(
                image_prompt,
                raw_message,
                response_channel=response_channel,
            )

        response_tasks = [
            task for task in [message_task, image_task] if task is not None
        ]
        await asyncio.wait(response_tasks)

    async def send_response(
        self,
        message: types.GenericMessage,
        raw_message: discord.Message,
        image_requested: bool,
    ) -> typing.Optional[typing.Tuple[asyncio.Task, discord.abc.Messageable]]:
        """
        Send a response to a message.

        Returns a tuple of the task that was created to send the message,
        and the channel that the message was sent to.

        If no message was sent, the task and channel will be None.
        """
        response_channel = raw_message.channel
        if (
            self.reply_in_thread
            and isinstance(raw_message.channel, discord.TextChannel)
            and isinstance(raw_message.author, discord.Member)
        ):
            # we want to create a response thread, if possible
            # but we have to see if the user has permission to do so
            # if the user can't we wont respond at all.
            perms = raw_message.channel.permissions_for(raw_message.author)
            if perms.create_public_threads:
                response_channel = await raw_message.create_thread(
                    name=f"{self.ai_name}, replying to "
                    + f"{raw_message.author.display_name}",
                )
                fancy_logger.get().debug(
                    f"Created response thread {response_channel.name} "
                    + f"({response_channel.id}) "
                    + f"in {raw_message.channel.name}"
                )
            else:
                # This user can't create threads, so we won't resond.
                # The reason we don't respond in the channel is that
                # it can create confusion later if a second user who
                # DOES have thread-create permission replies to that
                # message.  We'd end up creating a thread for that
                # second user's response, and again for a third user,
                # etc.
                fancy_logger.get().debug("User can't create threads, not responding.")
                return None

        response_coro = self.send_response_in_channel(
            message=message,
            raw_message=raw_message,
            image_requested=image_requested,
            response_channel=response_channel,
        )
        response_task = asyncio.create_task(response_coro)
        return (response_task, response_channel)

    async def history_plus_thread_kickoff_message(
        self,
        aiter: typing.AsyncIterator[discord.Message],
        limit: int,
    ) -> typing.AsyncIterator[types.GenericMessage]:
        """
        When returning the history of a thread, Discord
        does not include the message that kicked off the thread.

        It will show it in the UI as if it were, but it's not
        one of the messages returned by the history iterator.

        This method attempts to return that message as well,
        if we need it.
        """
        items = 0
        last_returned = None
        async for item in aiter:
            last_returned = item
            yield discord_utils.discord_message_to_generic_message(item)
            items += 1
        if last_returned is not None and items < limit:
            # we've reached the beginning of the history, but
            # still have space.  If this message was a reply
            # to another message, return that message as well.
            if last_returned.reference is not None:
                ref = last_returned.reference.resolved
                if ref is not None and isinstance(ref, discord.Message):
                    yield discord_utils.discord_message_to_generic_message(ref)

    async def recent_messages_following_thread(
        self, channel: discord.abc.Messageable
    ) -> typing.AsyncIterator[types.GenericMessage]:
        history = channel.history(limit=self.prompt_generator.history_lines)
        result = self.history_plus_thread_kickoff_message(
            history,
            limit=self.prompt_generator.history_lines,
        )
        return result

    async def send_response_in_channel(
        self,
        message: types.GenericMessage,
        raw_message: discord.Message,
        image_requested: bool,
        response_channel: discord.abc.Messageable,
    ) -> None:
        fancy_logger.get().debug(
            f"Request from {message.author_name} in {message.channel_name}"
        )

        recent_messages = await self.recent_messages_following_thread(response_channel)

        repeated_id = self.repetition_tracker.get_throttle_message_id(
            raw_message.channel.id
        )

        prompt_prefix = await self.prompt_generator.generate(
            ai_user_id=self.ai_user_id,
            message_history=recent_messages,
            image_requested=image_requested,
            throttle_message_id=repeated_id,
        )

        this_response_stat = self.response_stats.log_request_arrived(prompt_prefix)
        if self.log_all_the_things:
            print("prompt_prefix:\n----------\n")
            print(prompt_prefix)
            print("Response:\n----------\n")

        try:
            if self.dont_split_responses:
                generator = self.ooba_client.request_as_string(prompt_prefix)
            else:
                generator = self.ooba_client.request_by_sentence(prompt_prefix)

            async for sentence in generator:
                if self.log_all_the_things:
                    print(sentence)

                sentence = self.filter_immersion_breaking_lines(sentence)
                if not sentence:
                    # we can't send an empty message
                    continue

                response_message = await response_channel.send(sentence)
                generic_response_message = (
                    discord_utils.discord_message_to_generic_message(response_message)
                )
                self.repetition_tracker.log_message(
                    raw_message.channel.id, generic_response_message
                )

                this_response_stat.log_response_part()

        except Exception as err:
            fancy_logger.get().error(f"Error: {str(err)}")
            self.response_stats.log_response_failure()
            return

        this_response_stat.write_to_log(f"Response to {message.author_name} done!  ")
        self.response_stats.log_response_success(this_response_stat)

    def filter_immersion_breaking_lines(self, sentence: str) -> str:
        lines = sentence.split("\n")
        good_lines = []
        previous_line = ""
        for line in lines:
            # if the AI gives itself a second line, just ignore
            # the line instruction and continue
            if self.prompt_generator.bot_prompt_line == line:
                fancy_logger.get().warning(
                    f'Filtered out "{line}" from response, continuing'
                )
                continue

            # hack: abort response if it looks like the AI is
            # continuing the conversation as someone else
            if line.endswith(" says:"):
                fancy_logger.get().warning(
                    f'Filtered out "{line}" from response, aborting'
                )
                break

            if not line and not previous_line:
                # filter out multiple blank lines in a row
                continue

            good_lines.append(line)
        return "\n".join(good_lines)



import abc
import asyncio
import socket

import aiohttp


class OobaHttpClientError(Exception):
    pass


class SerializedHttpClient(abc.ABC):
    """
    Purpose: Limits the number of connections to a single host
    to one, so that we don't overwhelm the server.
    """

    HTTP_CLIENT_TIMEOUT_SECONDS: aiohttp.ClientTimeout = aiohttp.ClientTimeout(
        total=None,
        connect=None,
        sock_connect=5.0,
        sock_read=None,
    )

    @abc.abstractmethod
    async def _setup(self):
        # it's ok to raise an exception here, it will be caught
        ...

    async def setup(self):
        """
        Attempt to connect to the server.

        Returns:
            nothing, if the connection test was successful

        Raises:
            OobaHttpClientError, if the connection fails
        """
        try:
            await self._setup()
        except (
            aiohttp.ClientConnectionError,
            aiohttp.ClientError,
            ConnectionRefusedError,
            socket.gaierror,
            asyncio.exceptions.TimeoutError,
        ) as e:
            raise OobaHttpClientError(
                f"Could not connect to {self.service_name} server: [{self.base_url}]"
            ) from e

    def __init__(self, sevice_name: str, base_url: str):
        self.service_name = sevice_name
        self.base_url = base_url
        self._session = None

    def get_session(self) -> aiohttp.ClientSession:
        if not self._session:
            raise OobaHttpClientError("Session not initialized")
        return self._session

    async def __aenter__(self):
        connector = aiohttp.TCPConnector(limit_per_host=1)
        self._session = aiohttp.ClientSession(
            base_url=self.base_url,
            connector=connector,
            timeout=self.HTTP_CLIENT_TIMEOUT_SECONDS,
        )
        return self

    async def __aexit__(self, *_err):
        if self._session:
            await self._session.close()
        self._session = None

    def test_connection(self) -> None:
        async def try_setup():
            async with self:
                await self.setup()

        try:
            asyncio.run(try_setup())
        except AssertionError as e:
            # asyncio will throw an AssertionError if we try to run
            # with a base_url that has a path.  This is a user-supplied
            # value, so catching this is grody but necessary.
            raise OobaHttpClientError(
                f"Could not connect to {self.service_name} server: [{self.base_url}]"
            ) from e



import typing

from oobabot import fancy_logger
from oobabot import types


class RepetitionTracker:
    # how many times the bot can repeat the same thing before we
    # throttle history

    def __init__(self, repetition_threshold: int) -> None:
        self.repetition_threshold = repetition_threshold

        # stores a map of channel_id ->
        #   (last_message, throttle_message_id, repetion_count)

        self.repetition_count: typing.Dict[int, typing.Tuple[str, int, int]] = {}

    def get_throttle_message_id(self, channel_id: int) -> int:
        """
        Returns the message ID of the last message that should be throttled, or 0
        if no throttling is needed
        """
        _, throttle_message_id, _ = self.repetition_count.get(
            channel_id, (None, 0, None)
        )
        return throttle_message_id

    def log_message(
        self, channel_id: int, response_message: types.GenericMessage
    ) -> None:
        """
        Logs a message sent by the bot, to be used for repetition tracking
        """
        # make string into canonical form
        sentence = self.make_canonical(response_message.body_text)

        last_message, throttle_message_id, repetition_count = self.repetition_count.get(
            channel_id, ("", 0, 0)
        )
        if last_message == sentence:
            repetition_count += 1
        else:
            repetition_count = 0

        if repetition_count > 0:
            fancy_logger.get().debug(
                f"Repetition count for channel {channel_id} is {repetition_count}"
            )

        if self.should_throttle(repetition_count):
            fancy_logger.get().warning(
                "Repetition found, will throttle history for channel "
                + f"{channel_id} in next request"
            )
            throttle_message_id = response_message.message_id

        self.repetition_count[channel_id] = (
            sentence,
            throttle_message_id,
            repetition_count,
        )

    def hide_messages_before(self, channel_id: int, message_id: int) -> None:
        """
        Hides all messages before the given message ID in the given channel
        """
        sentence, _, repetition_count = self.repetition_count.get(
            channel_id, ("", 0, 0)
        )
        fancy_logger.get().info(
            "Hiding messages before message ID "
            + f"{message_id} in channel {channel_id}"
        )
        self.repetition_count[channel_id] = (sentence, message_id, repetition_count)

    def should_throttle(self, repetition_count: int) -> bool:
        """
        Returns whether the bot should throttle history for a given repetition count
        """
        return repetition_count >= self.repetition_threshold

    def make_canonical(self, content: str) -> str:
        return content.strip().lower()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#

import asyncio
import signal
import sys

from oobabot import bot_commands
from oobabot import decide_to_respond
from oobabot import discord_bot
from oobabot import fancy_logger
from oobabot import http_client
from oobabot import image_generator
from oobabot import ooba_client
from oobabot import prompt_generator
from oobabot import repetition_tracker
from oobabot import response_stats
from oobabot import sd_client
from oobabot import settings
from oobabot import templates


class OobaBot:
    def __init__(self, cli_args=None):
        fancy_logger.init_logging()

        self.settings = settings.Settings()
        self.settings.load(cli_args)
        if not self.settings.discord_token:
            msg = (
                f"Please set the '{self.settings.DISCORD_TOKEN_ENV_VAR}' "
                + "environment variable to your bot's discord token."
            )
            # will exit() after printing
            self.settings.error(msg)

        self.response_stats = None

        def sigint_handler(_signum, _frame):
            fancy_logger.get().info("Received SIGINT, exiting...")
            if self.response_stats is not None:
                self.response_stats.write_stat_summary_to_log()
            exit(1)

        signal.signal(signal.SIGINT, sigint_handler)

        ########################################################
        # Connect to Oobabooga

        self.ooba_client = ooba_client.OobaClient(
            self.settings.base_url, self.settings.OOBABOOGA_DEFAULT_REQUEST_PARAMS
        )

        ########################################################
        # Connect to Stable Diffusion, if configured

        self.stable_diffusion_client = None
        if self.settings.stable_diffusion_url:
            self.stable_diffusion_client = sd_client.StableDiffusionClient(
                base_url=self.settings.stable_diffusion_url,
                negative_prompt=self.settings.stable_diffusion_negative_prompt,
                negative_prompt_nsfw=self.settings.stable_diffusion_negative_prompt_nsfw,  # noqa: E501
                image_width=self.settings.image_width,
                image_height=self.settings.image_height,
                steps=self.settings.diffusion_steps,
                desired_sampler=self.settings.stable_diffusion_sampler,
            )

        ########################################################
        # Bot logic

        # decides which messages the bot will respond to
        self.decide_to_respond = decide_to_respond.DecideToRespond(
            self.settings.wakewords,
            self.settings.ignore_dms,
            self.settings.DECIDE_TO_RESPOND_INTERROBANG_BONUS,
            self.settings.DECIDE_TO_RESPOND_TIME_VS_RESPONSE_CHANCE,
        )

        # templates used to generate prompts to send to the AI
        # as well as for some UI elements
        self.template_store = templates.TemplateStore()

        # once we decide to respond, this generates a prompt
        # to send to the AI, given a message history
        self.prompt_generator = prompt_generator.PromptGenerator(
            ai_name=self.settings.ai_name,
            persona=self.settings.persona,
            history_lines=self.settings.history_lines,
            token_space=self.settings.OOBABOT_MAX_AI_TOKEN_SPACE,
            template_store=self.template_store,
            dont_split_responses=self.settings.dont_split_responses,
        )

        # tracks of the time spent on responding, success rate, etc.
        self.response_stats = response_stats.AggregateResponseStats(
            fn_get_total_tokens=lambda: self.ooba_client.total_response_tokens
        )

        # generates images, if stable diffusion is configured
        # also includes a UI to regenerate images on demand
        self.image_generator = None
        if self.stable_diffusion_client is not None:
            self.image_generator = image_generator.ImageGenerator(
                stable_diffusion_client=self.stable_diffusion_client,
                image_words=self.settings.image_words,
                template_store=self.template_store,
            )

        # if a bot sees itself repeating a message over and over,
        # it will keep doing so forever.  This attempts to fix that.
        # by looking for repeated responses, and deciding how far
        # back in history the bot can see.
        self.repetition_tracker = repetition_tracker.RepetitionTracker(
            repetition_threshold=self.settings.REPETITION_TRACKER_THRESHOLD
        )

        self.bot_commands = bot_commands.BotCommands(
            ai_name=self.settings.ai_name,
            decide_to_respond=self.decide_to_respond,
            repetition_tracker=self.repetition_tracker,
            reply_in_thread=self.settings.reply_in_thread,
            template_store=self.template_store,
        )

    def run(self):
        ########################################################
        # Test connection to services
        for client in [self.ooba_client, self.stable_diffusion_client]:
            if client is None:
                continue

            fancy_logger.get().info(f"{client.service_name} is at {client.base_url}")
            try:
                client.test_connection()
                fancy_logger.get().info(f"Connected to {client.service_name}!")
            except http_client.OobaHttpClientError as e:
                fancy_logger.get().error(
                    f"Could not connect to {client.service_name} "
                    + f"server: [{client.base_url}]"
                )
                fancy_logger.get().error("Please check the URL and try again.")
                if e.__cause__ is not None:
                    fancy_logger.get().error(f"Reason: {e.__cause__}")
                sys.exit(1)

        ########################################################
        # Connect to Discord

        fancy_logger.get().info("Connecting to Discord... ")
        bot = discord_bot.DiscordBot(
            self.ooba_client,
            decide_to_respond=self.decide_to_respond,
            prompt_generator=self.prompt_generator,
            repetition_tracker=self.repetition_tracker,
            response_stats=self.response_stats,
            image_generator=self.image_generator,
            bot_commands=self.bot_commands,
            ai_name=self.settings.ai_name,
            persona=self.settings.persona,
            ignore_dms=self.settings.ignore_dms,
            dont_split_responses=self.settings.dont_split_responses,
            reply_in_thread=self.settings.reply_in_thread,
            log_all_the_things=self.settings.log_all_the_things,
        )

        # opens http connections to our services,
        # then connects to Discord
        async def init_then_start():
            try:
                if self.stable_diffusion_client is not None:
                    async with self.stable_diffusion_client:
                        async with self.ooba_client:
                            try:
                                await bot.start(self.settings.discord_token)
                            finally:
                                await bot.close()
                else:
                    async with self.ooba_client:
                        try:
                            await bot.start(self.settings.discord_token)
                        finally:
                            await bot.close()
            except Exception as e:
                fancy_logger.get().error(f"Error starting bot: {e}")

        asyncio.run(init_then_start())


def main():
    oobabot = OobaBot()
    oobabot.run()



import time
import typing

from oobabot import fancy_logger


class ResponseStats:
    """
    Purpose: collects timing and rate statistics for a single response
    """

    def __init__(self, fn_get_total_tokens: typing.Callable[[], int], prompt: str):
        self.fn_get_total_tokens = fn_get_total_tokens
        self.start_time = time.time()
        self.start_tokens = fn_get_total_tokens()
        self.duration = 0
        self.latency = 0
        self.tokens = 0
        self.prompt_len = len(prompt)

    def log_response_part(self) -> None:
        """
        Call this each time the response is updated by the AI.
        """

        now = time.time()
        if not self.latency:
            self.latency = now - self.start_time
        self.duration = now - self.start_time
        self.tokens = self.fn_get_total_tokens() - self.start_tokens

    def tokens_per_second(self) -> float:
        """
        Returns the rate at which tokens were generated, in tokens per second.
        """
        if not self.duration:
            return 0
        return self.tokens / self.duration

    def write_to_log(self, log_prefix: str) -> None:
        """
        This writes the statistics for this specific
        request to the log.
        """
        fancy_logger.get().debug(
            log_prefix
            + f"tokens: {self.tokens}, "
            + f"time: {self.duration:.2f}s, "
            + f"latency: {self.latency:.2f}s, "
            + f"rate: {self.tokens_per_second():.2f} tok/s"
        )


class AggregateResponseStats:
    """
    Purpose: collects timing and rate statistics for all AggregateResponseStats
    """

    def __init__(self, fn_get_total_tokens: typing.Callable[[], int]):
        self.total_requests_received = 0
        self.total_successful_responses = 0
        self.total_failed_responses = 0
        self.total_response_time_seconds = 0
        self.total_response_latency_seconds = 0
        self.prompt_max_chars = 0
        self.prompt_min_chars = 0
        self.prompt_total_chars = 0
        self.fn_get_total_tokens = fn_get_total_tokens

    def log_request_arrived(self, prompt) -> ResponseStats:
        """
        Call this when a request has arrived.
        This must be followed by zero or more calls
        to log_response_part(), and then exactly one call to
        either log_response_failure() or log_response_success().
        """
        result = ResponseStats(self.fn_get_total_tokens, prompt)

        self.total_requests_received += 1
        # update the prompt stats now
        if self.prompt_max_chars < result.prompt_len:
            self.prompt_max_chars = result.prompt_len
        if self.prompt_min_chars > result.prompt_len:
            self.prompt_min_chars = result.prompt_len
        self.prompt_total_chars += result.prompt_len

        return result

    def log_response_failure(self) -> None:
        """
        Track the statistics for a failed response.
        """
        self.total_failed_responses += 1

    def log_response_success(self, response: ResponseStats) -> None:
        """
        Track the statistics for a successful response, and
        averages them into the overall statistics.

        Parameters:
         - response: the Response object returned
              by log_request_arrived()
        """
        self.total_successful_responses += 1
        self.total_response_time_seconds += response.duration
        self.total_response_latency_seconds += response.latency

    def error_rate(self) -> float:
        """
        Returns the percentage of requests that failed.
        """
        if 0 == self.total_requests_received:
            return 0.0
        return 100 * self.total_failed_responses / self.total_requests_received

    def average_response_time(self) -> float:
        """
        Returns the average response time in seconds.
        """
        if 0 == self.total_successful_responses:
            return 0.0
        return self.total_response_time_seconds / self.total_successful_responses

    def average_response_latency(self) -> float:
        """
        Returns the average response latency in seconds.
        """
        if 0 == self.total_successful_responses:
            return 0.0
        return self.total_response_latency_seconds / self.total_successful_responses

    def average_tokens_per_second(self) -> float:
        """
        Returns the average rate at which tokens were generated,
        in tokens per second.
        """
        if 0 == self.total_successful_responses:
            return 0.0
        return self.fn_get_total_tokens() / self.total_response_time_seconds

    def average_prompt_length(self) -> float:
        """
        Returns the average prompt length in characters.
        """
        if 0 == self.total_requests_received:
            return 0.0
        return self.prompt_total_chars / self.total_requests_received

    def write_stat_summary_to_log(self) -> None:
        """
        This writes a summary of the statistics to the log.
        Call this after all AggregateResponseStats have been handled.
        """
        if 0 == self.total_requests_received:
            fancy_logger.get().info("No requests handled")
            return

        fancy_logger.get().info(
            f"Recevied {self.total_requests_received} request(s), "
            + f"sent {self.total_successful_responses} successful responses "
            + f"and failed to send one {self.total_failed_responses} times(s)"
        )

        if self.total_failed_responses > 0:
            fancy_logger.get().error(
                "Error rate:                  " + f"{self.error_rate()}%"
            )

        if self.total_successful_responses > 0:
            fancy_logger.get().debug(
                "Average response time:       "
                + f"{self.average_response_time():6.2f}s"
            )
            fancy_logger.get().debug(
                "Average response latency:    "
                + f"{self.average_response_latency():6.2f}s"
            )
            fancy_logger.get().debug(
                "Average tokens per response: "
                + f"{self.average_tokens_per_second():6.2f}"
            )

        if self.total_response_time_seconds > 0:
            fancy_logger.get().debug(
                "Average tokens per second:   "
                + f"{self.average_tokens_per_second():6.2f}"
            )

        fancy_logger.get().debug(
            "Prompt length: "
            + f"max: {self.prompt_max_chars}, "
            + f"min: {self.prompt_min_chars}, "
            + f"avg: {self.average_prompt_length():.2f}"
        )



# Purpose: Streaming client for the Ooba API.
# Can provide the response by token or by sentence.
#

import typing

import aiohttp

from oobabot import fancy_logger
from oobabot import http_client
from oobabot import sentence_splitter


class OobaClient(http_client.SerializedHttpClient):
    # Purpose: Streaming client for the Ooba API.
    # Can provide the response by token or by sentence.

    SERVICE_NAME = "Oobabooga"

    OOBABOOGA_STREAMING_URI_PATH: str = "/api/v1/stream"

    END_OF_INPUT = ""

    def __init__(
        self, base_url: str, default_oobabooga_params: typing.Dict[str, typing.Any]
    ):
        super().__init__(self.SERVICE_NAME, base_url)
        self.total_response_tokens = 0
        self.default_oobabooga_params = default_oobabooga_params

    async def _setup(self):
        async with self.get_session().ws_connect(self.OOBABOOGA_STREAMING_URI_PATH):
            return

    async def request_by_sentence(self, prompt: str) -> typing.AsyncIterator[str]:
        """
        Yields each complete sentence of the response as it arrives.
        """

        splitter = sentence_splitter.SentenceSplitter()
        async for new_token in self.request_by_token(prompt):
            for sentence in splitter.by_sentence(new_token):
                yield sentence

    async def request_as_string(self, prompt: str) -> typing.AsyncIterator[str]:
        """
        Yields the entire response as a single string.
        """

        response = ""
        async for token in self.request_by_token(prompt):
            response += token
        yield response

    async def request_by_token(self, prompt: str) -> typing.AsyncIterator[str]:
        """
        Yields each token of the response as it arrives.
        """

        request: dict[
            str, typing.Union[bool, float, int, str, typing.List[typing.Any]]
        ] = {
            "prompt": prompt,
        }
        request.update(self.default_oobabooga_params)

        async with self.get_session().ws_connect(
            self.OOBABOOGA_STREAMING_URI_PATH
        ) as websocket:
            await websocket.send_json(request)

            async for msg in websocket:
                # we expect a series of text messages in JSON encoding,
                # like this:
                #
                # {"event": "text_stream", "message_num": 0, "text": ""}
                # {"event": "text_stream", "message_num": 1, "text": "Oh"}
                # {"event": "text_stream", "message_num": 2, "text": ","}
                # {"event": "text_stream", "message_num": 3, "text": " okay"}
                # {"event": "text_stream", "message_num": 4, "text": "."}
                # {"event": "stream_end", "message_num": 5}
                # get_logger().debug(f"Received message: {msg}")
                if msg.type == aiohttp.WSMsgType.TEXT:
                    # bdata = typing.cast(bytes, msg.data)
                    # get_logger().debug(f"Received data: {bdata}")

                    incoming_data = msg.json()
                    if "text_stream" == incoming_data["event"]:
                        self.total_response_tokens += 1
                        yield incoming_data["text"]

                    elif "stream_end" == incoming_data["event"]:
                        # Make sure any unprinted text is flushed.
                        yield self.END_OF_INPUT
                        return

                    else:
                        fancy_logger.get().warning(f"Unexpected event: {incoming_data}")

                elif msg.type == aiohttp.WSMsgType.ERROR:
                    fancy_logger.get().error(
                        f"WebSocket connection closed with error: {msg}"
                    )
                    raise http_client.OobaHttpClientError(
                        f"WebSocket connection closed with error {msg}"
                    )
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    fancy_logger.get().info(
                        f"WebSocket connection closed normally: {msg}"
                    )
                    return



import enum
import textwrap
import typing


class Templates(enum.Enum):
    COMMAND_LOBOTOMIZE_RESPONSE = "command_lobotomize_response"

    IMAGE_DETACH = "image_detach"
    IMAGE_CONFIRMATION = "image_confirmation"
    IMAGE_GENERATION_ERROR = "image_generation_error"
    IMAGE_UNAUTHORIZED = "image_unauthorized"

    PROMPT = "prompt"
    PROMPT_HISTORY_LINE = "prompt_history_line"
    PROMPT_IMAGE_COMING = "prompt_image_coming"


class TemplateToken(str, enum.Enum):
    AI_NAME = "AI_NAME"
    PERSONA = "PERSONA"
    IMAGE_COMING = "IMAGE_COMING"
    IMAGE_PROMPT = "IMAGE_PROMPT"
    MESSAGE_HISTORY = "MESSAGE_HISTORY"
    USER_MESSAGE = "USER_MESSAGE"
    USER_NAME = "USER_NAME"


class TemplateStore:
    # Purpose: store templates and format messages using them

    # mapping of template names to tokens allowed in that template
    TEMPLATES: typing.Dict[Templates, typing.List[TemplateToken]] = {
        Templates.COMMAND_LOBOTOMIZE_RESPONSE: [
            TemplateToken.AI_NAME,
            TemplateToken.USER_NAME,
        ],
        Templates.PROMPT: [
            TemplateToken.AI_NAME,
            TemplateToken.IMAGE_COMING,
            TemplateToken.MESSAGE_HISTORY,
            TemplateToken.PERSONA,
        ],
        Templates.PROMPT_HISTORY_LINE: [
            TemplateToken.USER_MESSAGE,
            TemplateToken.USER_NAME,
        ],
        Templates.PROMPT_IMAGE_COMING: [
            TemplateToken.AI_NAME,
        ],
        Templates.IMAGE_DETACH: [
            TemplateToken.IMAGE_PROMPT,
            TemplateToken.USER_NAME,
        ],
        Templates.IMAGE_CONFIRMATION: [
            TemplateToken.IMAGE_PROMPT,
            TemplateToken.USER_NAME,
        ],
        Templates.IMAGE_GENERATION_ERROR: [
            TemplateToken.IMAGE_PROMPT,
            TemplateToken.USER_NAME,
        ],
        Templates.IMAGE_UNAUTHORIZED: [TemplateToken.USER_NAME],
    }

    DEFAULT_TEMPLATES: typing.Dict[Templates, str] = {
        Templates.PROMPT: textwrap.dedent(
            """
            You are in a chat room with multiple participants.
            Below is a transcript of recent messages in the conversation.
            Write the next one to three messages that you would send in this
            conversation, from the point of view of the participant named
            {AI_NAME}.

            {PERSONA}

            All responses you write must be from the point of view of
            {AI_NAME}.
            ### Transcript:
            {MESSAGE_HISTORY}
            {IMAGE_COMING}
            """
        ),
        Templates.PROMPT_HISTORY_LINE: textwrap.dedent(
            """
            {USER_NAME} says:
            {USER_MESSAGE}

            """
        ),
        Templates.PROMPT_IMAGE_COMING: textwrap.dedent(
            """
            {AI_NAME}: is currently generating an image, as requested.
            """
        ),
        Templates.IMAGE_DETACH: textwrap.dedent(
            """
            {USER_NAME} tried to make an image with the prompt:
                '{IMAGE_PROMPT}'
            ...but couldn't find a suitable one.
            """
        ),
        Templates.IMAGE_CONFIRMATION: textwrap.dedent(
            """
            {USER_NAME}, is this what you wanted?
            If no choice is made, this message will 💣 self-destuct 💣 in 3 minutes.
            """
        ),
        Templates.IMAGE_GENERATION_ERROR: textwrap.dedent(
            """
            Something went wrong generating your image.  Sorry about that!
            """
        ),
        Templates.IMAGE_UNAUTHORIZED: textwrap.dedent(
            """
            Sorry, only {USER_NAME} can press the buttons.
            """
        ),
        Templates.COMMAND_LOBOTOMIZE_RESPONSE: textwrap.dedent(
            """
            Ummmm... what were we talking about?
            """
        ),
    }

    def __init__(self):
        self.templates: typing.Dict[Templates, TemplateMessageFormatter] = {}
        for template, tokens in self.TEMPLATES.items():
            template_fmt = TemplateStore.DEFAULT_TEMPLATES.get(template)
            if template_fmt is None:
                raise ValueError(f"Template {template} has no default format")
            self.add_template(template, template_fmt, tokens)

    def add_template(
        self,
        template_name: Templates,
        format_str: str,
        allowed_tokens: typing.List[TemplateToken],
    ):
        self.templates[template_name] = TemplateMessageFormatter(
            template_name, format_str, allowed_tokens
        )

    def format(
        self, template_name: Templates, format_args: typing.Dict[TemplateToken, str]
    ) -> str:
        return self.templates[template_name].format(format_args)


class TemplateMessageFormatter:
    # Purpose: format messages using a template string

    def __init__(
        self,
        template_name: Templates,
        template: str,
        allowed_tokens: typing.List[TemplateToken],
    ):
        self._validate_format_string(template_name, template, allowed_tokens)
        self.template_name = template_name
        self.template = template
        self.allowed_tokens = allowed_tokens

    def format(self, format_args: typing.Dict[TemplateToken, str]) -> str:
        return self.template.format(**format_args)

    @staticmethod
    def _validate_format_string(
        template_name: Templates,
        format_str: str,
        allowed_args: typing.List[TemplateToken],
    ):
        def find_all_ch(s: str, ch: str) -> typing.Generator[int, None, None]:
            # find all indices of ch in s
            for i, ltr in enumerate(s):
                if ltr == ch:
                    yield i

        # raises if fmt_string contains any args not in allowed_args
        allowed_close_brace_indices: typing.Set[int] = set()

        for open_brace_idx in find_all_ch(format_str, "{"):
            for allowed_arg in allowed_args:
                idx_end = open_brace_idx + len(allowed_arg) + 1
                next_substr = format_str[open_brace_idx : idx_end + 1]
                if next_substr == "{" + allowed_arg + "}":
                    allowed_close_brace_indices.add(idx_end)
                    break
            else:
                raise ValueError(
                    f"invalid template: {template_name} contains "
                    + f"an argument not in {allowed_args}"
                )
        for close_brace_idx in find_all_ch(format_str, "}"):
            if close_brace_idx not in allowed_close_brace_indices:
                raise ValueError(
                    f"invalid template: {template_name} contains "
                    + f"an argument not in {allowed_args}"
                )



# Purpose: Client for generating images using the AUTOMATIC1111
# API.  Takes a prompt and returns image binary data in PNG format.
#

import asyncio
import base64
import time
import typing

import aiohttp

from oobabot import fancy_logger
from oobabot import http_client

# todo: response stats for SD client


class StableDiffusionClient(http_client.SerializedHttpClient):
    """
    Purpose: Client for generating images using the AUTOMATIC1111
    API.  Takes a prompt and returns image binary data in PNG format.
    """

    SERVICE_NAME = "Stable Diffusion"
    LOG_PREFIX = SERVICE_NAME + ": "
    STABLE_DIFFUSION_API_URI_PATH: str = "/sdapi/v1/"

    API_COMMAND_URLS = {
        "get_samplers": STABLE_DIFFUSION_API_URI_PATH + "samplers",
        # get_samplers: GET only
        #   returns a list of samplers which we can use
        #   in text2img
        "options": STABLE_DIFFUSION_API_URI_PATH + "options",
        # options: GET and POST
        #   retrieves (GET) and sets (POST) global options
        #   for the server.  This is how we set the checkpoint
        #   and also a few privacy-related fields.
        "progress": STABLE_DIFFUSION_API_URI_PATH + "progress",
        # progress: GET only
        #   returns the progress of the current image generation
        "txt2img": STABLE_DIFFUSION_API_URI_PATH + "txt2img",
        # txt2img: POST only
        #   takes a prompt, generates an image and returns it
    }

    def __init__(
        self,
        base_url: str,
        negative_prompt: str,
        negative_prompt_nsfw: str,
        image_width: int,
        image_height: int,
        steps: int,
        desired_sampler: typing.Optional[str] = None,
    ):
        super().__init__(self.SERVICE_NAME, base_url)

        self.negative_prompt = negative_prompt
        self.negative_prompt_nsfw = negative_prompt_nsfw

        self._sampler = None
        self.desired_sampler = desired_sampler

        self.image_width = image_width
        self.image_height = image_height

        self._steps = steps

    # set default negative prompts to make it more difficult
    # to create content against the discord TOS
    # https://discord.com/guidelines

    DEFAULT_REQUEST_PARAMS: typing.Dict[str, typing.Union[bool, int, str]] = {
        # default values are commented out
        #
        # "do_not_save_samples": False,
        #    This is a privacy concern for the users of the service.
        #    We don't want to save the generated images anyway, since they
        #    are going to be on Discord.  Also, we don't want to use the
        #    disk space.
        "do_not_save_samples": True,
        #
        # "do_not_save_grid": False,
        #    Sames as above.
        "do_not_save_grid": True,
    }

    DEFAULT_OPTIONS = {
        #
        # - "enable_pnginfo"
        #   by default, this will store the parameters used to generate the image.
        #   For instance:
        #         Parameters: zombie taylor swift pinup movie posterSteps: 30,
        #           Sampler: DPM++ 2M Karras, CFG scale: 7, Seed: 1317257172,
        #           Size: 768x768, Model hash: 9aba26abdf, Model: Deliberate-2.0-15236
        #   We want to disable this, for a few reasons:
        #   - it's a privacy concern, since it's not clear to the user what
        #     is happening
        #   - it will include our negative prompts, which are actually bad
        #     phrases that represent content we DON'T want to generate.  If
        #     we upload these same strings to Discord, they might mistakenly
        #     flag the content as violating their TOS.  Even though the only
        #     reason those strings are used is to prevent that content.
        "enable_pnginfo": False,
        #
        # - "do_not_add_watermark"
        #    Similar to the above, we don't want to add a watermark to the
        #    generated image, just since it looks bad.  Also, it can leak
        #    the version of the SD software we are using, which isn't a good
        #    idea for security reasons.
        "do_not_add_watermark": True,
        #
        # - "samples_format"
        #   this is the format used for all image generation.
        "samples_format": "png",
    }

    async def set_options(self):
        url = self.API_COMMAND_URLS["options"]

        # first, get the current options.  We do this for two
        # reasons;
        #  - show the user exactly what we changed, if anything
        #  - some versions of the API don't support the
        #    "do_not_add_watermark" option, so we need to
        #    check if it's there before we try to set it
        #
        current_options = None
        async with self.get_session().get(url) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            current_options = await response.json()

        # now, set the options
        options_to_set = {}
        for k, v in self.DEFAULT_OPTIONS.items():
            if k not in current_options:
                continue
            if v == current_options[k]:
                continue
            options_to_set[k] = v
            fancy_logger.get().info(
                self.LOG_PREFIX
                + f" changing option '{k}' from to "
                + f"'{current_options[k]}' to '{v}'"
            )

        if not options_to_set:
            fancy_logger.get().debug(
                self.LOG_PREFIX + "Options are already set correctly, no changes made."
            )
            return

        async with self.get_session().post(url, json=options_to_set) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            await response.json()

    async def get_samplers(self) -> typing.List[str]:
        url = self.API_COMMAND_URLS["get_samplers"]
        async with self.get_session().get(url) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            response = await response.json()
            samplers = [str(sampler["name"]) for sampler in response]
            return samplers

    def generate_image(
        self,
        prompt: str,
        is_channel_nsfw: bool = False,
    ) -> "asyncio.Task[bytes]":
        # Purpose: Generate an image from a prompt.
        # Args:
        #     prompt: The prompt to generate an image from.
        #     negative_prompt: The negative prompt to use.
        #     sampler_name: The sampler to use.
        #     steps: The number of steps to use.
        #     width: The width of the image.
        #     height: The height of the image.
        # Returns:
        #     The image as bytes.
        # Raises:
        #     OobaHttpClientError, if the request fails.
        request = self.DEFAULT_REQUEST_PARAMS.copy()
        negative_prompt = self.negative_prompt
        if is_channel_nsfw:
            negative_prompt = self.negative_prompt_nsfw
        request.update(
            {
                "prompt": prompt,
                "negative_prompt": negative_prompt,
                "steps": self._steps,
                "width": self.image_width,
                "height": self.image_height,
            }
        )
        if self._sampler is not None:
            request["sampler_name"] = self._sampler

        async def do_post() -> bytes:
            fancy_logger.get().debug(
                self.LOG_PREFIX
                + f"Image request (nsfw: {is_channel_nsfw}): {request['prompt']}"
            )
            start_time = time.time()

            async with self.get_session().post(
                self.API_COMMAND_URLS["txt2img"],
                json=request,
            ) as response:
                if response.status != 200:
                    raise http_client.OobaHttpClientError(response)
                duration = time.time() - start_time
                json_body = await response.json()
                bytes = base64.b64decode(json_body["images"][0])
                fancy_logger.get().debug(
                    self.LOG_PREFIX
                    + "Image generated, {} bytes in {:.2f} seconds".format(
                        len(bytes), duration
                    )
                )
                return bytes

        # works around aiohttp being bad
        async def do_post_with_retry() -> bytes:
            tries = 0
            while True:
                try:
                    return await do_post()
                except aiohttp.ClientOSError as e:
                    if e.__cause__ is not ConnectionResetError:
                        raise e
                    if tries > 2:
                        raise e
                    fancy_logger.get().warning(
                        self.LOG_PREFIX
                        + f"Connection reset error: {e}, retrying in 1 second"
                    )
                    await asyncio.sleep(1)
                    tries += 1

        return asyncio.create_task(do_post_with_retry())

    async def set_sampler(self):
        """Sets the sampler to use, if it is available."""
        samplers = await self.get_samplers()
        if self.desired_sampler is not None:
            if self.desired_sampler in samplers:
                fancy_logger.get().debug(
                    self.LOG_PREFIX + "Using desired sampler '%s'", self.desired_sampler
                )
                self._sampler = self.desired_sampler
            else:
                fancy_logger.get().warn(
                    self.LOG_PREFIX + "Desired sampler '%s' not available",
                    self.desired_sampler,
                )
                self._sampler = None
        if self._sampler is None:
            fancy_logger.get().debug(
                self.LOG_PREFIX + "Using default sampler on SD server"
            )
            fancy_logger.get().debug(
                self.LOG_PREFIX + "Available samplers: %s", ", ".join(samplers)
            )
            self._sampler = None

    async def _setup(self):
        await self.set_sampler()
        await self.set_options()
        fancy_logger.get().debug(
            self.LOG_PREFIX
            + "Using negative prompt: "
            + self.negative_prompt[:20]
            + "..."
        )



import argparse
import os
import typing


class Settings(argparse.ArgumentParser):
    """
    User=customizable settings for the bot.  Reads from
    environment variables and command line arguments.
    """

    ############################################################
    # This section is for constants which are not yet
    # customizable by the user.

    # this is the number of tokens we reserve for the AI
    # to respond with.
    OOBABOT_MAX_NEW_TOKENS: int = 250

    # this is the number of tokens the AI has available
    # across its entire request + response
    OOBABOT_MAX_AI_TOKEN_SPACE: int = 730

    # This is a table of the probability that the bot will respond
    # in an unsolicited manner (i.e. it isn't specifically pinged)
    # to a message, based on how long ago it was pinged in that
    # same channel.
    DECIDE_TO_RESPOND_TIME_VS_RESPONSE_CHANCE: typing.List[
        typing.Tuple[float, float]
    ] = [
        # (seconds, base % chance of an unsolicited response)
        (60.0, 0.90),
        (120.0, 0.70),
        (60.0 * 5, 0.50),
    ]

    # increased chance of responding to a message if it ends with
    # a question mark or exclamation point
    DECIDE_TO_RESPOND_INTERROBANG_BONUS = 0.3

    # number of times in a row that the bot will repeat itself
    # before the repetition tracker will take action
    REPETITION_TRACKER_THRESHOLD = 1

    OOBABOOGA_DEFAULT_REQUEST_PARAMS: typing.Dict[
        str, typing.Union[bool, float, int, str, typing.List[typing.Any]]
    ] = {
        "max_new_tokens": OOBABOT_MAX_NEW_TOKENS,
        "do_sample": True,
        "temperature": 1.3,
        "top_p": 0.1,
        "typical_p": 1,
        "repetition_penalty": 1.18,
        "top_k": 40,
        "min_length": 0,
        "no_repeat_ngram_size": 0,
        "num_beams": 1,
        "penalty_alpha": 0,
        "length_penalty": 1,
        "early_stopping": False,
        "seed": -1,
        "add_bos_token": True,
        "truncation_length": OOBABOT_MAX_AI_TOKEN_SPACE,
        "ban_eos_token": False,
        "skip_special_tokens": True,
        "stopping_strings": [],
    }

    ############################################################
    # These are the default settings for the bot.  They can be
    # overridden by environment variables or command line arguments.

    # number lines back in the message history to include in the prompt
    DEFAULT_HISTORY_LINES_TO_SUPPLY = 7

    # words to look for in the prompt to indicate that the user
    # wants to generate an image
    DEFAULT_IMAGE_WORDS: typing.List[str] = [
        "drawing",
        "photo",
        "pic",
        "picture",
        "image",
        "sketch",
    ]

    # square image, 512x512
    DEFAULT_STABLE_DIFFUSION_IMAGE_SIZE: int = 512

    # 30 steps of diffusion
    DEFAULT_STABLE_DIFFUSION_STEPS: int = 30

    # ENVIRONMENT VARIABLES ####

    DISCORD_TOKEN_ENV_VAR: str = "DISCORD_TOKEN"
    DISCORD_TOKEN_ENV: str = os.environ.get(DISCORD_TOKEN_ENV_VAR, "")

    OOBABOT_PERSONA_ENV_VAR: str = "OOBABOT_PERSONA"
    OOBABOT_PERSONA: str = os.environ.get(OOBABOT_PERSONA_ENV_VAR, "")

    DEFAULT_WAKEWORDS: typing.List[str] = ["oobabot"]
    DEFAULT_URL: str = "ws://localhost:5005"

    # use this prompt for "age_restricted" Dsicord channels
    #  i.e. channel.nsfw is true
    DEFAULT_SD_NEGATIVE_PROMPT_NSFW: str = (
        "animal harm, "
        + "suicide, self-harm, "
        + "excessive violence, "
        + "naked children, child sexualization, lolicon"
    )

    # use this prompt for non-age-restricted channels
    DEFAULT_SD_NEGATIVE_PROMPT: str = (
        DEFAULT_SD_NEGATIVE_PROMPT_NSFW + ", sexually explicit content"
    )

    def __init__(self):
        self._settings = None
        self.wakewords = []
        super().__init__(
            description="Discord bot for oobabooga's text-generation-webui",
            epilog="Also, to authenticate to Discord, you must set the "
            + "environment variable:\n"
            f"\t{self.DISCORD_TOKEN_ENV_VAR} = <your bot's discord token>",
        )

        ###########################################################
        # Discord Settings

        discord_group = self.add_argument_group("Discord Settings")
        discord_group.add_argument(
            "--discord-token",
            type=str,
            default=self.DISCORD_TOKEN_ENV,
            help="Token to log into Discord with.  For security "
            + "purposes it's strongly recommended that you set "
            + f"this via the {self.DISCORD_TOKEN_ENV_VAR} environment "
            + "variable instead, if possible.",
        )
        discord_group.add_argument(
            "--dont-split-responses",
            default=False,
            help="If set, the bot post its entire response as a single "
            + "message, rather than splitting it into seperate "
            + "messages by sentence.",
            action="store_true",
        )
        discord_group.add_argument(
            "--history-lines",
            type=int,
            default=self.DEFAULT_HISTORY_LINES_TO_SUPPLY,
            help="Number of lines of chat history the AI will see "
            + "when generating a response.  The default is "
            + f"{self.DEFAULT_HISTORY_LINES_TO_SUPPLY}.",
        )
        discord_group.add_argument(
            "--ignore-dms",
            default=False,
            help="If set, the bot will ignore direct messages.",
            action="store_true",
        )
        discord_group.add_argument(
            "--wakewords",
            type=str,
            nargs="*",
            default=self.DEFAULT_WAKEWORDS,
            help="One or more words that the bot will listen for.\n "
            + "The bot will listen in all discord channels can "
            + "access for one of these words to be mentioned, then reply "
            + "to any messages it sees with a matching word.  "
            + "The bot will always reply to @-mentions and "
            + "direct messages, even if no wakewords are supplied.",
        )

        ###########################################################
        # Oobabooga Settings

        oobabooga_group = self.add_argument_group("Oobabooga Seetings")
        oobabooga_group.add_argument(
            "--ai-name",
            type=str,
            default="oobabot",
            help="Name of the AI to use for requests.  "
            + "This can be whatever you want, but might make sense "
            + "to be the name of the bot in Discord.",
        )
        oobabooga_group.add_argument(
            "--base-url",
            type=str,
            default=self.DEFAULT_URL,
            help="Base URL for the oobabooga instance.  "
            + "This should be ws://hostname[:port] for plain websocket "
            + "connections, or wss://hostname[:port] for websocket "
            + "connections over TLS.",
        )
        oobabooga_group.add_argument(
            "--log-all-the-things",
            default=False,
            help="Prints all oobabooga requests and responses in their "
            + "entirety to STDOUT",
            action="store_true",
        )
        oobabooga_group.add_argument(
            "--persona",
            type=str,
            default=self.OOBABOT_PERSONA,
            help="This prefix will be added in front of every user-supplied "
            + "request.  This is useful for setting up a 'character' for the "
            + "bot to play.  Alternatively, this can be set with the "
            + f"{self.OOBABOT_PERSONA_ENV_VAR} environment variable.",
        )
        oobabooga_group.add_argument(
            "--reply-in-thread",
            default=False,
            help="When set, the bot will generate a new thread for each "
            + "response it generates.  But it will only do so if the "
            + "user who prompted the bot has thread-create permissions.",
            action="store_true",
        )
        ###########################################################
        # Stable Diffusion Settings

        stable_diffusion_group = self.add_argument_group("Stable Diffusion Settings")
        stable_diffusion_group.add_argument(
            "--diffusion-steps",
            type=int,
            default=self.DEFAULT_STABLE_DIFFUSION_STEPS,
            help="Number of diffusion steps to take when generating an image.  "
            + f"The default is {self.DEFAULT_STABLE_DIFFUSION_STEPS}.",
        )
        stable_diffusion_group.add_argument(
            "--image-height",
            type=int,
            default=self.DEFAULT_STABLE_DIFFUSION_IMAGE_SIZE,
            help="Size of images to generate.  This is the height of the image "
            + "in pixels.  The default is "
            + f"{self.DEFAULT_STABLE_DIFFUSION_IMAGE_SIZE}.",
        )
        stable_diffusion_group.add_argument(
            "--image-width",
            type=int,
            default=self.DEFAULT_STABLE_DIFFUSION_IMAGE_SIZE,
            help="Size of images to generate.  This is the width of the image "
            + "in pixels.  The default is "
            + f"{self.DEFAULT_STABLE_DIFFUSION_IMAGE_SIZE}.",
        )
        stable_diffusion_group.add_argument(
            "--image-words",
            type=str,
            nargs="*",
            default=self.DEFAULT_IMAGE_WORDS,
            help="One or more words that will indicate the user "
            + "is requeting an image to be generated.",
        )
        stable_diffusion_group.add_argument(
            "--stable-diffusion-sampler",
            "--sd-sampler",
            type=str,
            default=None,
            help="Sampler to use when generating images.  If not specified, the one "
            + "set on the AUTOMATIC1111 server will be used.",
        )
        stable_diffusion_group.add_argument(
            "--stable-diffusion-url",
            "--sd-url",
            type=str,
            default=None,
            help="URL for an AUTOMATIC1111 Stable Diffusion server",
        )
        stable_diffusion_group.add_argument(
            "--sd-negative-prompt",
            type=str,
            default=self.DEFAULT_SD_NEGATIVE_PROMPT,
            help="Negative prompt to use when generating images.  This will discourage"
            + " Stable Diffusion from generating images with the specified content.  "
            + "By default, this is set to follow Discord's TOS.",
        )
        stable_diffusion_group.add_argument(
            "--sd-negative-prompt-nsfw",
            type=str,
            default=self.DEFAULT_SD_NEGATIVE_PROMPT_NSFW,
            help="Negative prompt to use when generating images in a channel marked as "
            + "'Age-Restricted'.",
        )

    def load(self, args) -> None:
        self._settings = self.parse_args(args=args).__dict__

        # Discord Settings
        self.discord_token = self._settings.pop("discord_token")
        self.dont_split_responses = self._settings.pop("dont_split_responses")
        self.history_lines = self._settings.pop("history_lines")
        self.ignore_dms = self._settings.pop("ignore_dms")
        self.reply_in_thread = self._settings.pop("reply_in_thread")
        self.wakewords = self._settings.pop("wakewords")

        # OogaBooga Settings
        self.ai_name = self._settings.pop("ai_name")
        self.base_url = self._settings.pop("base_url")
        print(self.base_url)
        self.log_all_the_things = self._settings.pop("log_all_the_things")
        self.persona = self._settings.pop("persona")

        # Stable Diffusion Settings
        self.diffusion_steps = self._settings.pop("diffusion_steps")
        self.image_height = self._settings.pop("image_height")
        self.image_width = self._settings.pop("image_width")
        self.image_words = self._settings.pop("image_words")
        self.stable_diffusion_negative_prompt = self._settings.pop("sd_negative_prompt")
        self.stable_diffusion_negative_prompt_nsfw = self._settings.pop(
            "sd_negative_prompt_nsfw"
        )
        self.stable_diffusion_sampler = self._settings.pop("stable_diffusion_sampler")
        self.stable_diffusion_url = self._settings.pop("stable_diffusion_url")

    def __repr__(self) -> str:
        return super().__repr__()
